// SPDX-FileCopyrightText: 2023 Jonah Brüchert <jbb@kaidan.im>
// SPDX-License-Identifier: GPL-2.0-or-later

#ifndef QMLTERMWIDGET_PLUGIN_H
#define QMLTERMWIDGET_PLUGIN_H

struct KonsoleQML {
    static void registerTypes(const char *uri);
};

#endif // QMLTERMWIDGET_PLUGIN_H
